open Syntaxe_abstraite
open Cmm
open Instr

(* trois tables de symboles. [lenv] pour les var. locales *)
(* [cenv] pour les var. de l'env de fermeture et *)
(* [genv] pour les var. globales *)
let lenv = Hashtbl.create 1
let cenv = Hashtbl.create 1
let genv = Hashtbl.create 1

let i = ref 0
let nouveau () = incr i; !i

let creer_etiquette () =
  let i = nouveau () in
  "etiq" ^ (string_of_int i)

(* tableau d'activation *)
(* application e1(e2)   *)
(*                *)
(* |      | <- fp *)
(* | v_e1 |       *)
(* | v_e2 |       *)
(* | res  |       *)
(* | ...  |       *)

let compile_var x =
  try
    (* variable introduite par un let *)
    [Pushl (Hashtbl.find lenv x)]
  with
    Not_found ->
      try
	(* variable de fermeture *)
	[Pushl (-1); Load (Hashtbl.find cenv x)]
      with Not_found ->
	(* dans ce cas, x est le parametre de la fonction *)
	[Pushl (-2)]
	  
(* stocker les variables libres dans l'env. de fermeture *)
let rec creer_env env indice =
  match env with
    [] -> []
  | x :: env ->
      (Dup 1) :: (compile_var x) @
      (Store indice :: (creer_env env (indice+1)))

let compile_clos s env =
  let d = ref 1 in
  List.iter (fun x -> Hashtbl.add cenv x !d; incr d) env;
  (* on stocke l'adresse de la fonction a l'indice 0 *)
  [Alloc (List.length env + 1);
   Dup 1; Pusha s; Store 0] @
  (creer_env env 1)  

(* nombre maximal de variables locales par bloc *)
let tfp_max = ref 0
let dg = ref 0

(* [dfp] est le deplacement relatif courant par rapport a fp *)
let rec compile tfp e =
  match e with
  | Mint i -> [(Pushi i)]
  | Mbool b -> if b then [Pushi 1] else [Pushi 0]
  | Mglobal x ->
      begin try
        [Pushg(Hashtbl.find genv x)]
      with Not_found -> failwith x end
  | Mlocal x ->
      compile_var x
  | Mop(o,e1,e2) -> 
      let op = 
        (* on traite quelques cas ici seulement *)
	match o with
          Eplus -> Add
        | Emoins -> Sub
        | Emult -> Mul
        | Ediv -> Div
      in
      (compile tfp e1) @ (compile tfp e2) @ [op]
  | Mlet(x, e1, e2) ->
      if !tfp_max = tfp then incr tfp_max;
      (* attention: ne marche que parce que [x] est unique *)
      (* dans le programme *)
      Hashtbl.add lenv x tfp;
      (compile tfp e1) @ (compile (tfp + 1) e2)
  | Mcall(e1, e2) ->
      (Pushn 1) :: (* place du resultat *)
      (compile tfp e2) @
      (compile tfp e1) @
      [Dup 1;Load 0;Call;Pop 2]
  | Mclos(s, env) ->
      compile_clos s env
  | Mifthenelse(e1, e2, e3) ->
      let faux = creer_etiquette () in
      let fin = creer_etiquette () in
      (compile tfp e1) @ 
      (Jz faux) :: (compile tfp e2) @
      [Jump fin; Label faux] @
      (compile tfp e3) @
      [Label fin]
	
(* code peu efficace ici *)
let prelude_decl d =
  match d with
    Mdef(s, e) ->
      let code = (compile 0 e) @ [Storeg !dg] in
      Hashtbl.add genv s !dg;
      incr dg;
      code
  | Mfundef _ -> []

let rec prelude_decl_list d_list =
  match d_list with
    [] -> []
  | d :: d_list ->
      let code = prelude_decl d in
      code @ (prelude_decl_list d_list)

let compile_code_decl d =
  match d with
    Mdef _ -> []
  | Mfundef(s, env, x, e) ->
      let index = ref 1 in
      List.iter
	(fun x -> Hashtbl.add cenv x !index; incr index) env;
      (Label s) :: (Pushn !tfp_max) :: (compile 0 e) @ [Return]
					
let rec compile_code_decl_list d_list = 
  match d_list with
    [] -> []
  | d :: d_list ->
      let code = compile_code_decl d in
      code @ (compile_code_decl_list d_list)

(* fonction principale *)
let compile decl_list =
  Pushn !dg ::
  (prelude_decl_list decl_list) @ (Stop :: (compile_code_decl_list decl_list))
  
